"""Time steppers"""
import attr
from toolz import merge, first
from . import tensordict
from .tensordict import TensorDict


@attr.s
class Batch(object):
    """A object for getting appropriate fields from a batch of data

    Attributes
    ----------
    data
    prognostics
    time_dim
    batch_dim
    """
    data = attr.ib(converter=TensorDict)
    prognostics = attr.ib()
    time_dim = 1
    batch_dim = 0

    @property
    def forcings(self):
        return set(self.data.keys()) - set(self.prognostics)

    @staticmethod
    def select_time(data, t):
        return data.apply(lambda x: x[:, t])

    def get_known_forcings(self):
        out = {}
        for key in self.prognostics:
            original_key = 'F' + key
            if original_key in self.data:
                out[key] = self.data[original_key]
        return TensorDict(out)

    def get_prognostics(self):
        return self.data[self.prognostics]

    def get_forcings_at_time(self, t):
        return self.select_time(self.data[self.forcings], t)

    def get_known_forcings_at_time(self, t):
        return self.select_time(self.get_known_forcings(), t)

    def get_prognostics_at_time(self, t):
        return self.select_time(self.data[self.prognostics], t)

    def get_model_inputs(self, t, prognostics=None):
        forcings = self.get_forcings_at_time(t)
        if prognostics is None:
            return TensorDict(merge(forcings, self.get_prognostics_at_time(t)))
        else:
            return TensorDict(merge(forcings, prognostics))

    @property
    def num_time(self):
        return self.data.size(self.time_dim)

    @property
    def size(self):
        """The batch size"""
        return self.data.size(self.batch_dim)

    def get_time_mean(self, key):
        return self.data[key].mean(self.time_dim)

    def data_for_lag(self, lag):
        return tensordict.lag(self.data[self.prognostics], lag, self.time_dim)


class TimeStepper:

    def __init__(self, source_function, time_step):
        self.source_function = source_function
        self.time_step = time_step

    def __call__(self, batch, initial_time=0, prediction_length=None):

        if prediction_length is None:
            prediction_length = batch.num_time - initial_time

        steps = predict_multiple_steps(
            self.source_function, batch, initial_time, prediction_length,
            self.time_step)
        state = [state for t, state, diags in steps]
        return tensordict.stack(state, dim=batch.time_dim)


def predict_multiple_steps(model, batch: Batch, initial_time,
                           prediction_length, time_step):
    """Yield Euler step predictions with a neural network"""
    state = batch.get_prognostics_at_time(initial_time)
    # yield initial_time, state, {}
    for t in range(initial_time, initial_time + prediction_length):
        inputs = batch.get_model_inputs(t, state)
        apparent_sources = model(inputs)
        known_forcing = batch.get_known_forcings_at_time(t)
        state = state + (apparent_sources/86400 + known_forcing) * time_step
        yield t + 1, state, apparent_sources


def select_keys_time(x, keys, t):
    return {key: x[key][:, t] for key in keys}



import logging

import joblib
import numpy as np
import json

from uwnet.thermo import compute_insolation

def initialize(state):
    pass


def finalize(state):
    pass


def get_configuration():
    try:
        with open("python_config.json") as f:
            return json.load(f)
    except FileNotFoundError:
        return {'models': []}



def get_model(config):
    type = config['type']
    if type=='sklearn_generic':
        model = joblib.load(config['path'])
        return model
    else:
        raise NotImplementedError(f"Model type {type} not implemented")


def state_format_to_sklearn_format(
        state,
        key_to_extract
):
    """
    takes the current state dict and returns single feature z profile np array input for sklearn_generic model
    """
    array_3d = state[key_to_extract]
    zdim, ydim, xdim = np.shape(array_3d)
    sklearn_format_matrix = array_3d \
        .transpose(2, 1, 0) \
        .reshape(xdim*ydim, zdim)
    return sklearn_format_matrix

def sklearn_format_to_state_format(
        model_output_array,
        state_arr_dims=(34, 64, 128)
):
    """
    assume the flattening order for input is y then x
    """
    zdim, ydim, xdim = state_arr_dims
    z_flattened_model = np.array(model_output_array).flatten()
    state_value = np.reshape(z_flattened_model , (zdim, ydim, xdim), 'F')
    return state_value


CONFIG = get_configuration()
MODELS =[get_model(model) for model in CONFIG['models']]


def call_sklearn_model(state):
    logger = logging.getLogger(__name__)
    # compute insolation. it is difficult to configure SAM to compute this
    # without including radiation
    state['SOLIN'] = compute_insolation(state['latitude'], state['day'])

    # get feature matrix out of state
    sl_sklearn_format = state_format_to_sklearn_format(
        state,
        key_to_extract='liquid_ice_static_energy'
    )
    qt_sklearn_format = state_format_to_sklearn_format(
        state,
        key_to_extract='total_water_mixing_ratio'
    )
    solin_sklearn_format = state_format_to_sklearn_format(
        state,
        key_to_extract='SOLIN'
    )
    sst_sklearn_format = state_format_to_sklearn_format(
        state,
        key_to_extract='sea_surface_temperature'
    )
    feature_matrix = np.hstack([
        sl_sklearn_format,
        qt_sklearn_format,
        solin_sklearn_format,
        sst_sklearn_format
    ])

    # get model output
    sklearn_model = MODELS[0]
    predicted_q1_q2 = sklearn_model.predict(feature_matrix)
    q1_state = sklearn_format_to_state_format(
        predicted_q1_q2[:, :34],
        state_arr_dims=(34, 64, 128)
    )
    q2_state = sklearn_format_to_state_format(
        predicted_q1_q2[:, 34:],
        state_arr_dims=(34, 64, 128)
    )
    # update keys 'tendency_of_...'
    state['tendency_of_liquid_ice_static_energy'] = q1_state
    state['tendency_of_total_water_mixing_ratio'] = q2_state


import torch
from toolz import merge_with, first, merge
import itertools
import dask.bag as db
import dask.array as da
from dask.delayed import delayed
import xarray as xr


def stack_dicts(delayeds):
    return merge_with(lambda x: torch.stack(x, dim=1), delayeds)


def select_time(batch, i):
    out = {}
    for key, val in batch.items():
        if val.dim() == 1:
            out[key] = val
        else:
            out[key] = val[i, :]
    return out


def get_batch_size(batch):
    return first(batch.values()).size(0)


def concat_dicts(delayeds, dim=1):
    return merge_with(lambda x: torch.cat(x, dim=dim), *delayeds)


def batch_to_model_inputs(batch, aux, prog, diag, forcing, constants):
    """Prepare batch for input into model

    This function reshapes the inputs from (batch, time, feat) to (time, z, x,
    y) and includes the constants

    """

    batch = merge(batch, constants)

    def redim(val, num):
        if num == 1:
            return val.t().unsqueeze(-1)
        else:
            return val.permute(1, 2, 0).unsqueeze(-1)

    # entries in inputs and outputs need to be tuples
    data_fields = set(map(tuple, aux + prog + diag))

    out = {key: redim(batch[key], num) for key, num in data_fields}

    for key, num in prog:
        if key in forcing:
            out['F' + key] = redim(batch['F' + key], num)

    return out


def centered_difference(x, dim=-3):
    """Compute centered difference with first order difference at edges"""
    x = x.transpose(dim, 0)
    dx = torch.cat(
        [
            torch.unsqueeze(x[1] - x[0], 0), x[2:] - x[:-2],
            torch.unsqueeze(x[-1] - x[-2], 0)
        ],
        dim=0)
    return dx.transpose(dim, 0)


def get_other_dims(x, dim):
    return set(range(x.dim())) - {dim}


def mean_over_dims(x, dims):
    """Take a mean over a list of dimensions keeping the as singletons"""
    for dim in dims:
        x = x.mean(dim=dim, keepdim=True)
    return x


def mean_other_dims(x, dim):
    """Take a mean over all dimensions but the one specified"""
    other_dims = get_other_dims(x, dim)
    return mean_over_dims(x, other_dims)


def split_by_chunks(dataset):
    chunk_slices = {}
    for dim, chunks in dataset.chunks.items():
        slices = []
        start = 0
        for chunk in chunks:
            stop = start + chunk
            slices.append(slice(start, stop))
            start = stop
        chunk_slices[dim] = slices
    for slices in itertools.product(*chunk_slices.values()):
        selection = dict(zip(chunk_slices.keys(), slices))
        yield delayed(lambda selection: dataset[selection].load())(selection)


def dataset_to_bag(dataset):
    return db.from_delayed(
        [delayed(lambda x: [x])(chunk) for chunk in split_by_chunks(dataset)])


def getarr(delayed_dataset, key, meta):
    delayed_dataset = delayed_dataset[0]
    delayed_arr = delayed_dataset[key]
    arr = da.from_delayed(delayed_arr, shape=meta.shape, dtype=meta.dtype)
    return xr.DataArray(arr, dims=meta.dims, coords=meta.coords)


def getds(delayed_dataset, meta):
    ds = xr.Dataset({})
    for key in meta:
        ds[key] = getarr(delayed_dataset, key, meta[key])

    return ds


def bag_to_dataset(bag, meta=None, dim='time'):
    if meta is None:
        meta = get_meta(bag)
    arrays = [getds(it, meta).drop(dim) for it in bag.to_delayed()]
    return xr.concat(arrays, dim)


def get_meta(bag):
    return bag.to_delayed()[0][0].compute()


def map_dataset(dataset, fun, dim):
    """Map a function accross the dimensions of a dataset

    Notes
    -----
    Computing a mean with the resulting dataset fails

    """
    chunked = dataset.chunk({dim: 1})
    # We don't really need to use a bag here. a list of delayed objects will do
    bag = dataset_to_bag(chunked)
    transformed = bag.map(fun)
    return bag_to_dataset(transformed)


def dataarray_to_broadcastable_array(
        array : xr.DataArray, dims):
    """Convert list an DataArray to an numpy array with a specified
    dimenionality

    Parameters
    ----------
    arrays: xr.DataArray
    dims : Sequence[str]
        desired ordering of the dimensions
    """
    # transpose data if necessary
    dims_in_array = [dim for dim in dims
                     if dim in array.dims]

    transposed = array.transpose(*dims_in_array)

    # expand null dims
    index = tuple(slice(None) if dim in array.dims else None
                  for dim in dims)

    # get data
    return transposed.data[index]


def dataset_to_broadcastable_array_dict(dataset, dims):
    return {key: dataarray_to_broadcastable_array(dataset[key], dims)
            for key in dataset.data_vars}



import pytest
import torch

from .groupby import DispatchByVariable, GroupBy


def test_dispatch_by_variables_membership():

    bins = [0, 1, 2, 3]
    expected = [0, 1, 3, 2, 4]
    a = torch.tensor(expected).float() - .5
    x = {'a': a.view(-1, 1)}
    model = DispatchByVariable(bins, bins, 'a', 0)
    membership = model.get_bin_membership(x)
    expected = [0, 1, 3, 2, 4]
    assert membership.tolist() == expected


@pytest.mark.parametrize('batch_shape', [(1, 3), (3, 1), (3, ), (1, 1, 3)])
def test_GroupBy_multiple_batch_dims(batch_shape):
    def fun(args):
        i, x = args
        return (i + 1) * x

    memberships = torch.tensor([0, 0, 1]).view(*batch_shape)
    x = torch.tensor([0, 1, 2]).view(*(*batch_shape, 1))
    out = GroupBy(memberships).apply(fun, x)

    assert out.shape == x.shape
    assert out.view(-1).tolist() == [0, 1, 4]


@pytest.mark.parametrize('memb', [
    [0, 1, 0],
    [1, 0, 0],
    [1, 2, 0],
])
def test_groupby_query(memb):
    memb = torch.tensor(memb)
    x = torch.tensor([
        0,
        1,
        2,
    ]).view(-1, 1)
    g = GroupBy(memb)
    y = torch.rand(3, 10, 3)
    out = g.apply(lambda x: y[x[0]].unsqueeze(0),
                  x)
    assert out.tolist() == y[memb].tolist()



import torch
from torch import nn
import math


def make_causal_mask(n, dependencies):
    dtype = torch.uint8

    cols = []
    for dependency in dependencies:
        col = torch.zeros(n, dtype=dtype)
        col[dependency] = 1
        cols.append(col)

    return torch.stack(cols, dim=1)


def make_causal_mask_from_ordinal(input, output, max=None, radius=1e-5):
    """Make a causal mask from an ordinal value

    Output j will have a value t^o(j) and input has value t^in(i). This output
    can depend only on inputs satisfying

    (t^in(i) <= t^o(j) AND t^in(i) < max) OR |t^in(i) - t^o(j)| < radius

    """
    mask = []
    input = input.float()
    output = output.float()
    if max is None:
        max = input[-1] + 10.0

    for val in output:
        col = (input <= val) * (input <= max)
        col[torch.abs(input - val) < radius] = 1
        mask.append(col)
    return torch.stack(mask, dim=1)


class CausalLinearBlock(nn.Module):
    """A linear block which forbids non-causal connections

    The user defines the notion of causality by ordering the inputs and outputs
    according to a list of real values. For example, one might assume that
    higher elevations cannot cause lower ones.

    Parameters
    ----------
    in_ordinal : torch.tensor
        has shape (*, n)
    out_ordinal : torch.tensor
        `(*, n)` This along with in_ordinal define the permitted connections
        within the weight matrix using :ref:`make_causal_mask_from_ordinal`.
        Has shape (*, m).
    activation
        The torch activation module to apply after. Defaults to ReLU.

    Returns
    -------
    out : torch.tensor
       Shape (*, m)
    """
    def __init__(self, in_ordinal, out_ordinal, activation=nn.ReLU()):
        super(CausalLinearBlock, self).__init__()

        self.activation = activation

        # setup mask
        self.register_buffer('in_ordinal', in_ordinal)
        self.register_buffer('out_ordinal', out_ordinal)
        mask = make_causal_mask_from_ordinal(in_ordinal, out_ordinal)
        self.register_buffer('mask', mask)

        # setup weights
        self._weight = nn.Parameter(
            torch.zeros(self.in_features, self.out_features))
        self.bias = nn.Parameter(torch.zeros(self.out_features))

        # intialize weights
        self.reset_parameters()

    @property
    def in_features(self):
        return len(self.in_ordinal)

    @property
    def out_features(self):
        return len(self.out_ordinal)

    @property
    def weight(self):
        zero = torch.zeros(1)
        return self._weight.where(self.mask, zero)

    def forward(self, x):
        y = torch.matmul(x, self.weight) + self.bias
        if self.activation is None:
            return y
        else:
            return self.activation(y)

    def reset_parameters(self):
        n = self.out_features
        stdv = 1. / math.sqrt(n)
        self._weight.data.uniform_(-stdv, stdv)
        self.bias.data.uniform_(-stdv, stdv)



from functools import partial

import numpy as np
from toolz import curry, pipe

import torch

from .tensordict import TensorDict


def _numpy_dict_to_torch_dict(inputs):
    return TensorDict(
        {key: torch.from_numpy(val)
         for key, val in inputs.items()})


def _torch_dict_to_numpy_dict(output):
    return {key: val.numpy() for key, val in output.items()}


def call_with_numpy_dict(self, inputs, **kwargs):
    """Call the neural network with numpy inputs"""
    with torch.no_grad():
        return pipe(inputs, _numpy_dict_to_torch_dict,
                    partial(self, **kwargs), _torch_dict_to_numpy_dict)


@curry
def NumpyWrapper(model, state):

    # Pre-process the inputs
    kwargs = {}
    for key, val in state.items():
        if isinstance(val, np.ndarray):
            kwargs[key] = val

    return call_with_numpy_dict(model, kwargs)



from functools import partial

import numpy as np
import pytest
import torch
from pytest import approx
from .timestepper import Batch, TensorDict

from .loss import (compute_multiple_step_loss, r2_score,
                   weighted_mean_squared_error, weighted_r2_score,
                   multiple_step_loss_step)
from uwnet.utils import mean_over_dims, mean_other_dims
from unittest.mock import Mock

approx = partial(approx, abs=1e-6)


def _time_stepping_problem():
    n = 10
    shape = (1, n)

    def criterion(x, y):
        return torch.abs(x - y).mean()

    def model(x):
        return TensorDict({'x': torch.zeros(1, requires_grad=True).float()})

    prognostics = ['x']
    batch = TensorDict(
        {'x': torch.zeros(shape).float(), 'Fx': torch.zeros(shape).float()})
    batch = Batch(batch, prognostics)
    return batch, model, criterion, n


def test_compute_multiple_step_loss():
    batch, model, criterion, n = _time_stepping_problem()
    loss = compute_multiple_step_loss(
        criterion, model, batch, 0, n - 1, 1.0)
    assert loss.item() == pytest.approx(0.0)


# @pytest.mark.skip()
def test_multiple_step():
    batch, model, criterion, n = _time_stepping_problem()

    self = Mock()
    self.model = model
    self.time_step = 1.0
    engine = Mock()
    multiple_step_loss_step(self, engine, batch)



@pytest.mark.parametrize('x,w,dim,expected', [
    (torch.rand(10), torch.rand(10), -1, 0.0),
    (torch.rand(10, 5), torch.rand(5), -1, 0.0),
    (torch.rand(5, 10), torch.rand(5), 0, 0.0),
    (torch.rand(4, 5, 6), torch.rand(5), 1, 0.0),
])
def test_weighted_mean_squared_error(x, w, dim, expected):
    loss = weighted_mean_squared_error(x, x, w, dim)
    assert loss.dim() == 0
    assert loss.item() == pytest.approx(expected)


def test_weighted_mean_squared_error_value():
    """Test the MSE function for a non-zero value"""
    a = 2
    x = torch.rand(5)
    y = x + a
    w = torch.rand(5)
    expected = a ** 2 * w.mean()
    loss = weighted_mean_squared_error(x, y, w)
    assert loss.item() == pytest.approx(expected.item())


def test_mean_over_dims():
    a = torch.rand(10)

    # single dimension
    res = mean_over_dims(a, [0])
    np.testing.assert_allclose(res.item(), a.mean().item())

    # multiple dimensions
    a = torch.rand(10, 2)
    res = mean_over_dims(a, [0, 1])
    assert res.item() == approx(a.mean().item())

    # incomplete reduction
    a = torch.rand(3, 4, 5)
    res = mean_over_dims(a, [0, 1])
    assert res.shape == (1, 1, 5)


def test_mean_other_dims():
    a = torch.rand(3, 4, 5)
    mu = mean_other_dims(a, 1)
    assert mu.shape == (1, 4, 1)


def test_weighted_r2_score():
    a = torch.rand(10, 2)
    w = torch.tensor([1.0, 1.0])

    score = weighted_r2_score(a, a, w, dim=-1)
    assert score.item() == approx(1.0)

    score = weighted_r2_score(a, a.mean(), w, dim=-1)
    assert score.item() == approx(0.0)


def test_r2_score():
    a = torch.rand(10, 2)

    score = r2_score(a, a)
    assert score.item() == approx(1.0)

    score = r2_score(a, a.mean())
    assert score.item() == approx(0.0)



import numpy as np
import pytest
import torch
from toolz import curry
from .testing import assert_tensors_allclose, mock_data
from .normalization import Scaler
from uwnet.tensordict import TensorDict

approx = curry(pytest.approx, abs=1e-6)


def test_scaler():
    x = torch.rand(10)
    mean = x.mean()
    scale = x.std()

    scaler = Scaler({'x': mean}, {'x': scale})
    y = scaler({'x': x})
    scaled = y['x']
    assert scaled.mean().item() == approx(0.0)
    assert scaled.std().item() == approx(1.0)


def test_scaler_fit_xarray():
    name = 'a'
    ds = mock_data(init=np.random.random).to_dataset(name=name)
    scaler = Scaler().fit_xarray(ds)

    expected = torch.from_numpy(ds.groupby('z').mean(...)[name].values)
    assert_tensors_allclose(scaler.mean[name], expected)


def test_scaler_fit_generator():
    shape = (10, 3, 4, 1, 1)
    name = 'a'
    a = torch.rand(shape)

    def generator():
        for arr in a.split(2, dim=0):
            yield TensorDict({name: a})

    scaler = Scaler().fit_generator(generator())

    expected_mean = a.mean(0).mean(0).squeeze()
    assert_tensors_allclose(scaler.mean[name], expected_mean)



from unittest.mock import Mock

import numpy as np
import pytest
import torch

import xarray as xr

from .xarray_interface import (
    _array_to_tensor, call_with_xr, _torch_dict_to_dataset)


def init_dataarray(shape):
    dims = ['time', 'z', 'y', 'x']
    coords = {dim: np.arange(n) for dim, n in zip(dims, shape)}
    return xr.DataArray(np.ones(shape), dims=dims, coords=coords)


def test__array_to_tensor_3d_input():
    t, z, y, x = shape = (3, 10, 2, 2)
    arr = init_dataarray(shape)
    out = _array_to_tensor(arr)
    assert out.shape == shape

    # test for 3D data
    two_d_data = arr.isel(z=0)
    out = _array_to_tensor(two_d_data)
    assert out.shape == (t, 1, y, x)


def test_call_with_xr_empty_data_fails():
    shape = (1, 1, 0, 1)
    name = 'a'

    da = init_dataarray(shape).to_dataset(name=name)
    mock = Mock()
    with pytest.raises(
            ValueError, match="'y' dimension of 'a' has length 0"):
        call_with_xr(mock, da)


def test_call_with_xr_no_error():
    shape = (10, 1, 1, 1)
    name = 'a'


    da = init_dataarray(shape).to_dataset(name=name)

    # setup mock object
    mock = Mock()
    a = torch.tensor(da[name].values)
    mock.return_value = {name: a}

    call_with_xr(mock, da)


def test__torch_dict_to_dataset():
    name = 'a'
    shape = (10, 1, 1, 1)
    da = init_dataarray(shape).to_dataset(name=name)
    torch_dict = {name: torch.tensor(da[name].values)}
    _torch_dict_to_dataset(torch_dict, da.coords)



import os
import numpy as np
import xarray as xr
from . import ndimage_xarray
from .datasets import tiltwave



def test_ndimage():
    a = tiltwave()
    a.ndimage

    # try gaussian filter
    a.ndimage.gaussian_filter(dict(x=1.0, z=0.0))

    # try one dimensional filter
    a.ndimage.gaussian_filter1d('x', 1.0)



"""Objects for sampling along xarray objects
"""
import xarray as xr
import numpy as np


class Sampler:
    """Samples along a set of dimensions of an xarray object"""

    def __init__(self,
                 input_variables,
                 output_variables,
                 feature_dims,
                 sample_dims,
                 feature_valid_fn=None):
        self.inputs = input_variables
        self.outputs = output_variables
        self.feature_dims = feature_dims
        self.sample_dims = sample_dims
        self.feature_valid_fn = (feature_valid_fn
                                 if feature_valid_fn else feature_valid)

    def get_valid_features(self, features):
        return [self.feature_valid_fn(var, z) for (var, z) in features]

    def to_input_data_matrix(self, data):
        ds = data.stack(samples=self.sample_dims)
        x = ds[self.inputs].to_stacked_array('feature', self.feature_dims)
        valid_feats = self.get_valid_features(self.feature.values)
        return x[:, valid_feats]

    def to_output_vector(self, data):
        ds = data.stack(sample=self.sample_dims)[self.outputs]
        return ds

    def from_output_vector(self, data, samples):
        return xr.DataArray(
            data, dims='samples', coords={'samples':
                                          samples}).unstack('samples')

    def prepare_data(self, data):
        return self.to_input_data_matrix(data), self.to_output_vector(data)


class SklearnWrapper:
    def __init__(self, sampler, model):
        self.sampler = sampler
        self.model = model

    def predict(self, data):
        x = self.sampler.to_input_data_matrix(data)
        y = self.model.predict(x)
        return self.sampler.from_output_vector(y, x.samples)

    def fit(self, data):
        x, y = self.sampler.prepare_data(data)
        [x_train, y_train], [x_test, y_test] = split_train_test(x, y)
        self.model.fit(x_train, y_train)

        print("Training score", self.model.score(x_train, y_train))
        print("Testing score", self.model.score(x_test, y_test))
        return self


def feature_valid(var, z):
    if var == 'QT':
        if z < 10000:
            return True
        else:
            return False
    if var == 'SLI':
        if z < 10000:
            return True
        else:
            return False
    return True


def split_train_test(x, y):

    nsample = 10000

    valid_feats = [feature_valid(var, z) for (var, z) in x.feature.values]

    inds = np.random.choice(x.shape[0], nsample)
    inds.sort()
    x_train, y_train = x[inds, valid_feats].values, y[inds].values

    inds = np.random.choice(x.shape[0], nsample)
    inds.sort()
    x_test, y_test = x[inds, valid_feats].values, y[inds].values

    return [x_train, y_train], [x_test, y_test]



import xarray as xr
import json
import argparse
from src.sam.case import get_ngqaua_ic
from uwnet.data.blur import blur_dataset
from uwnet.thermo import layer_mass
from src.data.ngaqua import NGAqua
from src.sam.process_ngaqua import run_sam_nsteps
from subprocess import run
import os
import shutil
from os.path import abspath


class Namespace:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def get_args_snakemake():
    print("Using snakemake")
    sigma = snakemake.params.sigma
    if sigma:
        sigma = float(sigma)

    return Namespace(
        time_step=int(snakemake.wildcards.step),
        sam_parameters=snakemake.input.sam_parameters,
        sam=snakemake.config.get('sam_path',  '/opt/sam'),
        ngaqua_root=snakemake.params.ngaqua_root,
        output=snakemake.output[0],
        sigma=sigma)


def get_args_argparse():

    parser = argparse.ArgumentParser(
        description='Pre-process a single time step')
    parser.add_argument('-n', '--ngaqua-root', type=str)
    parser.add_argument('-s', '--sam', type=str, default='/opt/sam')
    parser.add_argument('-t', '--time-step', type=int, default=0)
    parser.add_argument('-p', '--sam-parameters', type=str, default=0)
    parser.add_argument('--sigma', type=float, help='Radius for Gaussian blurring')

    parser.add_argument('output')
    return parser.parse_args()


def get_args():
    """Get the arguments needed to run this script

    This will function will behave differently if run from snakemake"""

    try:
        snakemake
    except NameError:
        return get_args_argparse()
    else:
        return get_args_snakemake()


args = get_args()


def get_extra_features(ngaqua: NGAqua, time_step):

    # compute layer_mass
    stat = ngaqua.stat
    rho = stat.RHO.isel(time=0).drop('time')
    w = layer_mass(rho)

    # get 2D variables
    time = ngaqua.times_3d[time_step]
    d2 = ngaqua.data_2d.sel(time=time)

    # add variables to three-d
    d2['RADTOA'] = d2.LWNT - d2.SWNT
    d2['RADSFC'] = d2.LWNS - d2.SWNS
    d2['layer_mass'] = w
    d2['rho'] = rho

    # 3d variables
    qrad = ngaqua.data_3d.QRAD.drop(['x', 'y'])
    d2['QRAD'] = qrad.sel(time=time)

    return d2


# get initial condition
ic = get_ngqaua_ic(args.ngaqua_root, args.time_step)

# get data
ngaqua = NGAqua(args.ngaqua_root)
features = get_extra_features(ngaqua, args.time_step)

if args.sigma:
    print(f"Blurring data with radius {args.sigma}")
    ic = blur_dataset(ic, args.sigma)
    features = blur_dataset(features, args.sigma)

# compute the forcings by running through sam
if args.sam_parameters:
    with open(args.sam_parameters) as f:
        prm = json.load(f)

path = run_sam_nsteps(ic, prm, sam_src=abspath(args.sam))
files = os.path.join(path, 'OUT_3D', '*.nc')
ds = xr.open_mfdataset(files)
shutil.rmtree(path)

for key in ['QT', 'SLI', 'U', 'V']:
    forcing_key = 'F' + key
    src = ds[key].diff('time') / ds.time.diff('time') / 86400
    src = src.isel(time=0)
    ds[forcing_key] = src

ds = ds.isel(time=0)

# forcing data
ic['x'] = ds.x
ic['y'] = ds.y
for key in ic.data_vars:
    if key not in ds:
        ds[key] = ic[key]

ds = ds.merge(features).expand_dims('time')
ds.attrs['sam_namelist'] = json.dumps(prm)
ds.to_netcdf(args.output, unlimited_dims=['time'], engine='h5netcdf')



import json
from json import JSONEncoder
import numpy as np

class NumpyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyArrayEncoder, self).default(obj)


def dump(arg, f):
    """Serialize a linearized response function dictionary"""
    return json.dump(arg, f, cls=NumpyArrayEncoder)


def lists_to_ndarray(d):
    if isinstance(d, dict):
        return {key: lists_to_ndarray(d[key]) for key in d}
    elif isinstance(d, list):
        return np.array(d)
    else:
        return d

def load(f):
    """Deserialize a linearized response function from file"""
    loaded = json.load(f)
    return lists_to_ndarray(loaded)





import click
from dataclasses import dataclass, field
import joblib
import json
import logging
import os
import xarray as xr

# current has import error and does not recognize XarrayBatchLoader as importable
from .datasets_handler import XarrayLoaderSklearn
from uwnet.thermo import sec_in_day

# test with rf regressor first


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('train_sklearn_model')




@dataclass
class ModelTrainingConfig:
    model_params: dict
    model_type: str

    data: str = "data/processed/reshaped/noBlur.zarr"
    existing_data_file: str = "data/processed/generic_sklearn_format/"
    predict_radiation: bool = True
    seed: int = 1234
    vertical_grid_size: int = 34
    num_samples: int = 20000
    num_samples_validation: int = 10000
    batch_size: int = 256
    num_batches_for_training: int = 5
    prognostics: list = field(default_factory=lambda: ['QT', 'SLI'])
    targets: list = field(default_factory=lambda: ['Q1', 'Q2'])

    def __post_init__(self):
        self.training_slices = slice(0, self.num_samples)
        self.validation_slices = slice(self.num_samples, self.num_samples + self.num_samples_validation)
        self.output_path = f"sklearn_models/{self.model_type}.pkl"

def get_model_training_config(model_config_path):
    with open(model_config_path, 'r') as f:
        config_dict = json.load(f)
    model_config = ModelTrainingConfig(**config_dict)
    return model_config


def get_dataset(config):
    logger.info(f"Opening xarray dataset {config.data}")
    try:
        dataset = xr.open_zarr(config.data)
    except ValueError:
        dataset = xr.open_dataset(config.data)

    if not config.predict_radiation:
        dataset['FSLI'] = dataset['FSLI'] + dataset['QRAD'] / sec_in_day

    try:
        return dataset.isel(step=0).drop('step').drop('p')
    except:
        return dataset


def get_data_loader(
        data: xr.Dataset,
        config,
        train
):
    if train:
        slices = config.training_slices
    else:
        slices = config.validation_slices
    data = data.isel(x=slices)

    # List needed variables
    variables = config.prognostics + ['SST', 'SOLIN', 'QRAD']
    for variable in config.prognostics:
        forcing_key = 'F' + variable
        variables.append(forcing_key)
    train_data = XarrayLoaderSklearn(
        dataset=data,
        batch_size=config.batch_size,
        variables=variables)
    return train_data



def get_regressor_model(model_config):
    if 'rf' in model_config.model_type or 'random_forest' in model_config.model_type:
        from sklearn.ensemble import RandomForestRegressor
        regressor = RandomForestRegressor(**model_config.model_params)
        # temporarily set n_estimators to 0, will add estimators for each batch in iteration
        regressor.n_estimators = 0
    elif 'xgb' in model_config.model_type and 'tree' in model_config.model_type:
        # TODO: add other options?
        #from xgboost.sklearn_generic import XGBRegressor
        pass
    else:
        logger.error('model_type {self.model_type} provided in config file is not a valid model type.')

    return regressor

@click.command()
@click.option(
    '-cf',
    '--model-config-file',
    type=click.Path(),
    help='Location of config for model training.')
@click.option(
    '-op',
    '--model-output-path',
    type=click.Path(),
    default=None,
    help='Output the trained model to this path.')
def main(
        model_output_path,
        model_config_file,
):
    model_training_config = get_model_training_config(model_config_file)
    if not model_output_path:
        model_output_path = model_training_config.output_path
    dataset = get_dataset(model_training_config)
    train_loader = get_data_loader(dataset, model_training_config, train=True)
    regressor = get_regressor_model(model_training_config)

    for batch_num, training_batch in enumerate(train_loader):
        if batch_num >= model_training_config.num_batches_for_training:
            break
        train_X, train_y = training_batch
        if model_training_config.num_samples:
            train_X = train_X[:model_training_config.num_samples]
            train_y = train_y[:model_training_config.num_samples]
        logger.info(('training batch {0} / {1}'.format(batch_num, model_training_config.num_batches_for_training)))
        if 'rf' in model_training_config.model_type or 'random_forest' in model_training_config.model_type:
            regressor.n_estimators += model_training_config.model_params['n_estimators']

        regressor.fit(train_X, train_y)
    model_save_filename = os.path.join(
        model_output_path, f'{model_training_config.model_type}.pkl')
    joblib.dump(regressor, model_save_filename)


if __name__ == '__main__':
    main()



"""
Energy and moisture are conserved if the following vertically integrated quantities are true:

    cp <Q1> = L_v Prec + SHF + RADSFC - RADTOP
    <Q2> = Evap - Prec

"""
import torch
from toolz import curry
import torch.nn.functional as F


def mass_integrate(x, w):
    return (x * w).sum(-1, keepdim=True)


def apply_linear_constraint(lin, a, x, *args, inequality=False, v=None,
                            **kwargs):
    """Apply a linear constraint

    Parameters
    ----------
    lin : Callable
        the linear constraint of the form lin(x, *args, **kwargs) = a that is linear in x
    inequality : bool
        assume that lin(x, *args, **kwargs) >= a

    Returns
    -------
    y : Tensor
      x transformed to satisfy the constraint

    """

    val_x = lin(x, *args, **kwargs)

    # use a constant adjustment
    # x - alpha * v
    if v is None:
        v = torch.ones(x.size(-1))
    val_v = lin(v, *args, **kwargs)
    alpha = (val_x - a)

    if inequality:
        alpha = alpha.clamp(max=0)

    return x - v * (alpha / val_v)


@curry
def expected_moisture(q0, fqt, precip, lhf, h, layer_mass):
    """Same as energy imbalance but for moisture budget

    Parameters
    ----------
    q0  (g/kg)
        moisture before
    fqt (g/kg/s)
        moistening from horizontal advection.
    h   (s)
        time step
    precip (mm/d)
        sfc flux over time step
    lhf  (W/m2)

    Returns
    -------
    current_pw, expected_pw (g/m2)

    """
    Lv = 2.51e6
    density_liquid = 1000.0

    water0 = mass_integrate(q0, layer_mass) / 1000.0  # kg
    fqt_int = mass_integrate(fqt, layer_mass) / 1000.0  # kg/s
    net_evap = lhf / Lv - precip / 1000.0 * density_liquid / 86400  # kg/s
    return water0 * 1000, 1000 * (water0 + (fqt_int + net_evap) * h)


def precip_to_energy(prec):
    """Convert precip to W/m2

    Parameters
    ---------
    prec : (mm/day)
    """
    density_water = 1000
    Lv = 2.51e6
    coef = density_water / 1000 / 86400 * Lv
    return coef * prec


def expected_temperature(sl, fsl, prec, shf, radtoa, radsfc, h, layer_mass):
    """Expected column average SLI to conserve energy

    Parameters
    ----------
    sl (K)
        output before neural network
    fsl (K/s)
        heat transport without surface fluxes
    prec (mm/day)
        change in precipitable water over time step
    shf, radtoa, radsfc (W/m2)
        surface and toa fields (upward)
    h (day)
        time step
    layer_mass (kg/m2)

    Returns
    -------
    sl0, sl_int (K / kg / m^2)
        integrals of before/after SLI

    """
    cp = 1004

    sl0 = mass_integrate(layer_mass, sl)
    fsl_int = mass_integrate(fsl, layer_mass)

    energy_rate_of_change = precip_to_energy(prec) + shf + radsfc - radtoa
    sl1 = sl0 + h * (fsl_int + energy_rate_of_change / cp)
    return sl0, sl1


def enforce_expected_integral(x, int, layer_mass):
    int_actual = mass_integrate(x, layer_mass)
    return x * int / int_actual


def fix_negative_moisture(q, layer_mass):
    """Fix negative moisture in moisture conserving way"""
    eps = torch.tensor(1e-10, requires_grad=False)

    water_before = mass_integrate(q, layer_mass)
    q = q.clamp(eps)
    water_after = mass_integrate(q, layer_mass)
    return q * water_before / water_after


def apply_constraints(x0, x1, time_step, output_specs):
    """Main function for applying constraints"""

    layer_mass = x0['layer_mass']
    qt_name = 'QT'
    sl_name = 'SLI'
    fsl_name = 'FSLI'
    fqt_name = 'FQT'

    for spec in output_specs:
        if spec.positive and (spec.name in x1):
            if spec.conserved:
                x1[spec.name] = fix_negative_moisture(x1[spec.name],
                                                      x0['layer_mass'])
            else:
                x1[spec.name] = F.softplus(x1[spec.name])

    if 'Prec' in x1:
        pw0, pw = expected_moisture(x0[qt_name], x0[fqt_name], x1['Prec'],
                                    x1['LHF'], time_step, x0['layer_mass'])

        sl_int0, sl_int = expected_temperature(
            x0[sl_name], x0[fsl_name], x1['Prec'], x1['SHF'], x1['RADTOA'],
            x1['RADSFC'], time_step, layer_mass)

        sl = enforce_expected_integral(x1[sl_name], sl_int, layer_mass)
        qt = enforce_expected_integral(x1[qt_name], pw, layer_mass)
    else:
        sl = x1[sl_name]
        qt = x1[qt_name]

    x1[qt_name] = qt
    x1[sl_name] = sl
    return x1



import pytest
import torch

from .causal import (CausalLinearBlock, make_causal_mask,
                     make_causal_mask_from_ordinal)
from .testing import assert_tensors_allclose


def test_CausalLinearBlock():
    in_ordering = torch.tensor([1, 2, 3])
    out_ordering = torch.tensor([1, 1, 2, 2, 3, 3, 3])
    linear = CausalLinearBlock(in_ordering, out_ordering, activation=None)

    assert linear.in_features == len(in_ordering)
    assert linear.out_features == len(out_ordering)

    linear.weight

    x = torch.rand(1, 3, requires_grad=True)
    y = linear(x)
    assert y.size(-1) == len(out_ordering)

    # test dependency structure of module
    y[0, 0].backward()
    grad = x.grad[0, :]
    assert grad[-1].item() == pytest.approx(0)
    assert grad[0].item() != pytest.approx(0)

    x = torch.rand(1, 3, requires_grad=True)
    y = linear(x)
    y[0, 2].backward()
    grad = x.grad[0, :]
    assert grad[-1].item() == pytest.approx(0)
    assert grad[0].item() != pytest.approx(0)
    assert grad[1].item() != pytest.approx(0)


def test_make_causal_mask():
    n = 2
    dependency = [[0, 1], [0]]
    expected = torch.tensor([[1, 1], [1, 0]])

    out = make_causal_mask(n, dependency)
    assert_tensors_allclose(out, expected)


@pytest.mark.parametrize('input,output,expected,m', [
    ([0, 1], [0, 1], [[1, 1], [0, 1]], None),
    ([0, 1.0], [1.0, 1.0], [[1, 1], [1, 1]], None),
    ([1.0, 1.0], [1.0], [[1], [1]], .5),
    ([0.9, 1.0], [1.0], [[0], [1]], .5),
    ([0.4, 1.0], [1.0], [[1], [1]], .5),
])
def test_make_causal_mask_from_ordinal(input, output, expected, m):
    input = torch.tensor(input)
    output = torch.tensor(output)
    mask = make_causal_mask_from_ordinal(input, output, m)

    expected = torch.tensor(expected)

    assert_tensors_allclose(mask, expected)



"""Tools for pre/post processing inputs/outputs from neural networks

Unlike pytorch's builtin tools, this code allows building pytorch modules from
scikit-learn estimators.
"""
import pandas as pd
import numpy as np
import xarray as xr
import torch
from torch import nn

from uwnet.thermo import compute_apparent_source
from uwnet.modules import MapByKey, LinearFixed
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

import logging

logger = logging.getLogger(__name__)


def prepare_data(data, exog=['QT', 'SLI', 'SST', 'SOLIN'], sample=None):
    """Flatten XArray dataset into dataframe.

    The columns of the dataframe are the variables names and height indices.
    The rows are the samples, which are taken along 'x', 'y', and 'time'.
    """
    logger.info("Flattening xarray Dataset to pandas DataFrame")
    size = data[exog].nbytes/1e9
    logger.info(f"Size: {size} GB")
    vals = []
    names = []
    sample_dims = ['x', 'y', 'time']
    for key in exog:
        val = data[key].stack(s=sample_dims)
        if 'z' not in val.dims:
            val = val.expand_dims('z')

        arr = val.transpose('s', 'z').values
        vals.append(arr)

        for z in range(arr.shape[1]):
            names.append((key, z))

    inputs = np.concatenate(vals, axis=1)
    idx = pd.MultiIndex.from_tuples(names)
    return pd.DataFrame(inputs, columns=idx)


def fit_pca_inverse_transform(q1, n=16):
    pca = PCA(n_components=n, whiten=True)
    pca.fit(q1)
    return LinearFixed.from_affine(pca.inverse_transform, n)


def fit_pca_transform(x, n=16):
    pca = PCA(n_components=n, whiten=True)
    pca.fit(x)
    return LinearFixed.from_affine(pca.transform, x.shape[1])


def fit_scaler_transform(x):
    scaler = StandardScaler().fit(x)
    return LinearFixed.from_affine(scaler.transform, x.shape[1])


def get_post(data, m=20):
    logger.info("Fitting PCA models for outputs")
    q1 = compute_apparent_source(data.SLI, data.FSLI * 86400).dropna('time')
    q2 = compute_apparent_source(data.QT, data.FQT * 86400).dropna('time')

    ds = xr.Dataset({'Q1': q1, 'Q2': q2})
    df = prepare_data(ds, exog=['Q1', 'Q2'])

    funcs = {
        'SLI': fit_pca_inverse_transform(df['Q1'], m),
        'QT': fit_pca_inverse_transform(df['Q2'], m),
    }

    return MapByKey(funcs)


def get_pre(data, m=20):
    logger.info("Building preprocessing module")
    keys = ['QT', 'SLI', 'SHF', 'LHF', 'SOLIN', 'SST']
    df = prepare_data(data, exog=keys)
    transformers = {}
    for key in keys:
        x = df[key]
        n = x.shape[1]

        if n == 1:
            logger.info(f"Fitting Scaler for {key}")
            transformers[key] = fit_scaler_transform(x)
        else:
            logger.info(f"Fitting PCA for {key}")
            transformers[key] = fit_pca_transform(x, m)

    return MapByKey(transformers)



class Post(nn.Module):

    def __init__(self, q0, inputs):
        super(Post, self).__init__()
        self.register_buffer('q0', q0)
        self.inputs = inputs

    def forward(self, d):
        q0 = self.q0.clamp(max=1)
        d['QT'] = d['QT'] * q0
        return d


class LowerAtmosInput(nn.Module):

    @property
    def outputs(self):
        return [
            ('QT', 15),
            ('SLI', 18),
            ('SST', 1),
            ('SOLIN', 1),
        ]

    def forward(self, x):
        out = {}
        for key, n in self.outputs:
            out[key] = x[key][..., :n]
        return out


class IdentityOutput(nn.Module):
    inputs = [
        ('QT', 34),
        ('SLI', 34),
    ]

    outputs = [
        ('QT', 34),
        ('SLI', 34),
    ]

    def forward(self, x):
        return x


class Sequential(nn.Sequential):
    @property
    def inputs(self):
        return self[0].inputs

    @property
    def outputs(self):
        return self[-1].outputs


def get_pre_post_orig(data, data_loader, n):
    from .normalization import Scaler
    inputs = [('QT', n), ('SLI', n),
              ('SST', 1), ('SOLIN', 1)]
    scaler = Scaler().fit_generator(data_loader)
    scaler.outputs = inputs
    scaler.inputs = inputs
    # post processor
    outputs = (('QT', n), ('SLI', n))
    post = Post(scaler.scale['QT'], outputs)
    return scaler, post


def get_pre_post(data, data_loader, _config):
    kind = _config['kind']
    logger.info(f"Getting pre/post processor of type {kind}")

    if kind == 'pca':
        return get_pre(data, m=20), get_post(data, m=20)
    elif kind == 'saved':
        path = _config['path']
        logger.info(f"Loading pre/post module from {path}")
        return torch.load(path)
    elif kind == 'lower_atmos':
        pre, post = get_pre_post_orig(data, data_loader, n=34)
        lower = LowerAtmosInput()
        pre = Sequential(pre, lower)
        return pre, post
    elif kind == 'orig':
        return get_pre_post_orig(data, data_loader, n=34)
    elif kind == 'identity':
        return LowerAtmosInput(), IdentityOutput()
    else:
        raise NotImplementedError



"""Train neural network parametrizations using the single or
multiple step loss function

This code uses a tool called Sacred_ to log important details about each
execution of this script to a database, and to specify hyper parameters of the
scheme (e.g. learning rate).

Examples
--------

This script can be executed with the following command::

   python -m uwnet.train with data=<path to dataset>

To see a list of all the available configuration options run::

   python -m uwnet.train print_config

.. _Sacred: https://github.com/IDSIA/sacred

"""
import matplotlib
matplotlib.use('agg')
import logging
import os
from contextlib import contextmanager
from os.path import join

import xarray as xr
from sacred import Experiment
from toolz import curry

import torch
from torch.optim.lr_scheduler import StepLR
from ignite.engine import Engine, Events
from .datasets_handler import get_timestep, XarrayBatchLoader, get_dataset, get_data_loader
from uwnet.loss import get_input_output, get_step, weighted_mean_squared_error
from uwnet.model import get_model
from uwnet.pre_post import get_pre_post
from uwnet.training_plots import TrainingPlotManager
from uwnet.metrics import WeightedMeanSquaredError

ex = Experiment("Q1", interactive=True)

TrainingPlotManager = ex.capture(TrainingPlotManager, prefix='plots')

get_model = ex.capture(get_model, prefix='model')
get_pre_post = ex.capture(get_pre_post, prefix='prepost')
get_dataset = ex.capture(get_dataset)
get_data_loader = ex.capture(get_data_loader)


@ex.config
def my_config():
    """Default configurations managed by sacred"""
    train_data = ""
    test_data = ""

    predict_radiation = True
    lr = .001
    epochs = 2
    model_dir = 'models'
    skip = 5
    time_length = None
    batch_size = 256
    vertical_grid_size = 34
    loss_scale = {
        'LHF': 150,
        'SHF': 10,
        'RADTOA': 600.0,
        'RADSFC': 600.0,
        'U': 5.0,
        'V': 5.0,
        'Prec': 8,
        'QP': 0.05,
        'QN': 0.05,
        'QT': 1.0,
        'SLI': 2.5
    }

    output_dir = None

    prognostics = ['QT', 'SLI']
    prepost = dict(kind='pca', path='models/prepost.pkl')

    model = dict(kind='inner_model')

    plots = dict(interval=1, single_column_locations=[(32, 0)])
    step = dict(
        name='instability',
        kwargs={'alpha': 1.0}
    )
    lr_decay_rate = None
    lr_step_size = 5


def get_plot_manager(model):
    from src.data import open_data
    dataset = open_data('training')
    return TrainingPlotManager(ex, model, dataset)


def get_validation_engine(model, dt):
    def _validate(engine, data):
        from uwnet.timestepper import Batch
        # TODO this code duplicaates loss.py:107
        batch = Batch(data.float(), prognostics=['QT', 'SLI'])
        return get_input_output(model, dt, batch)

    return Engine(_validate)


@ex.capture
def get_output_dir(_run=None, model_dir=None, output_dir=None):
    """Get a unique output directory name using the run ID that sacred
    assigned OR return the specified output directory
    """
    if output_dir:
        return output_dir
    else:
        file_name = str(_run._id)
        return join(model_dir, file_name)


class Trainer(object):
    """Utility object for training a neural network parametrization
    """

    @ex.capture
    def __init__(self, _run, lr, loss_scale, train_data, test_data,
                 lr_decay_rate=None, lr_step_size=5):
        # setup logging
        logging.basicConfig(level=logging.INFO)

        self.compute_metrics = False

        self.logger = logging.getLogger(__name__)

        # get output directory
        self.output_dir = get_output_dir()

        train_dataset = get_dataset(train_data)
        test_dataset = get_dataset(test_data)

        self.mass = torch.tensor(train_dataset.layer_mass.values).view(
            -1, 1, 1).float()
        self.z = torch.tensor(train_dataset.z.values).float()
        self.time_step = get_timestep(train_dataset)
        self.train_loader = get_data_loader(train_dataset)
        self.test_loader = get_data_loader(test_dataset)
        self.model = get_model(*get_pre_post(train_dataset, self.train_loader))
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr)

        if lr_decay_rate is None:
            self.lr_scheduler = None
        else:
            self.lr_scheduler = StepLR(
                self.optimizer, step_size=lr_step_size, gamma=lr_decay_rate)

        self.criterion = weighted_mean_squared_error(
            weights=self.mass / self.mass.mean(), dim=-3)
        self.plot_manager = get_plot_manager(self.model)
        self.setup_validation_engine()
        self.setup_engine()

    def step_lr_scheduler(self):
        if self.lr_scheduler is not None:
            self.lr_scheduler.step()

    def log_validation_results(self, trainer):
        self.tester.run(self.test_loader)
        metrics = self.tester.state.metrics

        log_str = "test metrics: "
        for name, val in metrics.items():
            ex.log_scalar("train_" + name, val)
            log_str += f'{name}: {val:.2f}\t'
        self.logger.info(log_str)

        metrics = trainer.state.metrics
        log_str = "train metrics: "
        for name, val in metrics.items():
            ex.log_scalar("test_" + name, val)
            log_str += f'{name}: {val:.2f}\t'
        self.logger.info(log_str)

    def setup_validation_engine(self):
        self.tester = get_validation_engine(self.model, self.time_step)
        self.setup_metrics_for_engine(self.tester)

    @ex.capture(prefix='step')
    def get_step(self, name, kwargs, _log):
        _log.info(f"Using `{name}` gradient stepper")
        if name == 'multi':
            self.logger.info(f"Disabling computation of metrics")
            self.compute_metrics = False
        else:
            self.compute_metrics = True
        return get_step(name, self, kwargs)

    def setup_engine(self):
        step = self.get_step()
        self.engine = Engine(step)
        self.setup_metrics_for_engine(self.engine)

        self.engine.add_event_handler(Events.ITERATION_COMPLETED,
                                      self.after_batch)
        self.engine.add_event_handler(Events.ITERATION_COMPLETED,
                                      self.print_loss_info)
        self.engine.add_event_handler(Events.EPOCH_COMPLETED, self.after_epoch)
        self.engine.add_event_handler(Events.EPOCH_COMPLETED,
                                      self.log_validation_results)

    @ex.capture
    def setup_metrics_for_engine(self, engine, prognostics):

        if not self.compute_metrics:
            return

        @curry
        def output_transform(key, args):
            x, y = args
            return x[key], y[key]

        for key in prognostics:
            metric = WeightedMeanSquaredError(
                self.mass, output_transform=output_transform(key))

            metric.attach(engine, key)

    def print_loss_info(self, engine):
        n = len(self.train_loader)
        batch = engine.state.iteration % (n + 1)
        log_str = f"[{batch}/{n}]:\t"
        for key, val in engine.state.loss_info.items():
            log_str += f'{key}: {val:.2f}\t'
            ex.log_scalar(key, val)
        self.logger.info(log_str)

    def after_batch(self, engine):
        state = engine.state
        batch_info = {
            'epoch': state.epoch,
            'loss': state.output,
        }
        ex.log_scalar('loss', batch_info['loss'])

    def after_epoch(self, engine):
        # save artifacts
        n = engine.state.epoch
        epoch_file = f"{n}.pkl"

        with self.change_to_work_dir():
            torch.save(self.model, epoch_file)
            ex.add_artifact(epoch_file)
            self.plot_manager(engine)

        self.step_lr_scheduler()

    def _make_work_dir(self):
        try:
            os.makedirs(self.output_dir)
        except OSError:
            pass

    @contextmanager
    def change_to_work_dir(self):
        """Context manager for using a working directory"""
        self.logger.info(f"Saving outputs in {self.output_dir}")
        self._make_work_dir()
        try:
            cwd = os.getcwd()
            os.chdir(self.output_dir)
            yield
        finally:
            os.chdir(cwd)

    @ex.capture
    def train(self, epochs):
        """Train the neural network for a fixed number of epochs"""
        self.engine.run(self.train_loader, max_epochs=epochs)


@ex.command()
def train_pre_post(prepost):
    """Train the pre and post processing modules"""
    dataset = get_dataset()
    logging.info(f"Saving Pre/Post module to {prepost['path']}")
    torch.save(get_pre_post(dataset), prepost['path'])


@ex.automain
def main():
    Trainer().train()



import numpy as np
import pytest
import torch
from pytest import approx
from torch.nn import Module

import xarray as xr
from uwnet.xarray_interface import call_with_xr
from uwnet.modules import MOE

sl_name = 'SLI'


def _assert_all_close(x, y):
    np.testing.assert_allclose(y.detach().numpy(), x.detach().numpy())


def _mock_batch(t, z, y, x, init=torch.rand):
    return {
        'LHF': init(t, y, x),
        'SHF': init(t, y, x),
        'SOLIN': init(t, y, x),
        'QT': init(t, z, y, x),
        'SLI': init(t, z, y, x),
        'FQT': init(t, z, y, x),
        'FSLI': init(t, z, y, x),
        'layer_mass': torch.arange(1, z + 1).float(),
        # 'p': init(nz),
    }


def test_moe():

    m = 10
    n = 5
    n_exp = 3

    x = torch.rand(100, m)

    rhs = MOE(m, n, n_exp)
    out = rhs(x)
    return out.size() == (100, n)


@pytest.mark.xfail()
def test_stepper_no_cheating():
    """Make sure that stepper only uses the initial point"""
    # TODO modify this test for new timesteppers

    n = 10
    z = 34
    batch = _mock_batch(10, z, 10, 10)
    stepper = _stepper(z)

    # mock out the forward method
    class MockStep(Module):
        def forward(self, x):
            return {key: val + 1.0 for key, val in x.items()}

    stepper.rhs = MockStep()

    for key, val in batch.items():
        if key not in {'FSLI', 'FQT'}:
            val.requires_grad = True

    pred = stepper(batch)

    # backprop
    sl = pred['SLI']
    sl[-1, 0, 0, 0].backward()

    sl_true = batch['SLI']
    # the gradient of sl_true should only be nonzero for the first n points
    # of the dat
    grad_t = (sl_true.grad**2).mean(-1).mean(-1).mean(-1)
    assert torch.sum(grad_t[n:]).item() == approx(0)
    assert torch.sum(grad_t[0:n]).item() > 0


@pytest.mark.xfail()
def test_ApparentSource():
    def dummy_mod(x):
        return x

    mod = ApparentSource(dummy_mod, inputs=(('a', 34), ), outputs=(('a', 34),
    ))

    batch = {'a': torch.ones(128, 34, 1, 1)}
    y = mod(batch)
    a = y['a']
    assert a.shape == batch['a'].shape


@pytest.mark.skip()
@pytest.mark.parametrize('drop_time', [0, 1, 2])
def test_call_with_xr(drop_time):
    def model(d):
        return {key: val[drop_time:] for key, val in d.items()}

    dims = ['time', 'z', 'y', 'x']
    shape = [10, 11, 12, 13]
    coords = {dim: np.arange(n) for dim, n in zip(dims, shape)}

    ds = xr.Dataset({'a': (dims, np.random.rand(*shape))}, coords=coords)
    out = call_with_xr(model, ds, drop_times=drop_time)
    np.testing.assert_array_equal(ds.a.values[drop_time:], out.a.values)



from .wave import WaveCoupler, LinearResponseFunction, ablate_upper_atmosphere, WaveEq
from .spectra import compute_spectrum, scatter_spectra
from .serialize import load, dump
from .. import thermo
from .. import jacobian
